// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: CalCoreData
// Author: Hassan Siddiqui
//
// Description: Data/methods for ARF generation specific to EPIC CAL data
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: CalCoreData.h,v 1.6 2014/04/03 13:37:51 rsaxton Exp $

#ifndef CalCoreData_H
#define CalCoreData_H

/* General */

/* SAS-specific classes */
#include "Xmm.h"
#include "CalCommon.h"
#include "CalServer.h"
#include "Tracer.h"

/* Local classes */
#include "CoreData.h"
#include "GeneralIterator.h"
#include "SourcePosition.h"
#include "CamCoords.h"
#include "TelCoords.h"
#include "PixCoords.h"
#include "PositionUtils.h"

class CalCoreData : public CoreData
{
public:

  /** Constructors  */
  CalCoreData( double psfenergy,   
	       const string & datamode,
	       const string & dateobs,
	       const PosWCS * const posxy
	       );

  /**  Destructors  */
  ~CalCoreData();

  /* Change state of other accessors which depend on ccd number. */
  bool setCcd(const CamCoords & camcoords);

  /* Ebins */
  const BinnedAxis * ebins() const;

 /******************************************************
   * 
   * Geometry based routines
   *
   ******************************************************/

  const SourcePosition * sourcePosition() const;

  void setSourcePosition( const string & sourceCoordType,
                      double x, double y);

  const PositionUtils * positionUtils() const;
  PositionUtils * varpositionUtils();

  PsfModel *      psf()        const;

  /* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
  const CamCoords detxy005ToCamCoord2( int detx_005, 
				       int dety_005 ) const;

  /* Convert from CAMCOORD2 to TELCOORD */
  const TelCoords camCoord2ToTelCoord( const CamCoords & camcoords ) const;

  /* Convert from CAMCOORD2 to PIXCOORD */
  const PixCoords camCoord2ToPixCoord( const CamCoords & camcoords ) const;

  /* Convert from X/Y to DETX/DETY */
  SourcePosition xyToDetxy (double x, double y) const;

  /* Frame period, in seconds */
  double framePeriod()          const;


  /** Iterator methods. */
  GeneralIterator<int> * ccdIterator();

  /** Data resource methods */

  EffectiveAreaCurve effectiveArea  () const;

  EffectiveAreaCurve effectiveArea  ( double theta, double phi ) const;

  EffectiveAreaCurve mirrorAdjustment(EffectiveAreaCurve eain) const;

  EffectiveAreaCurve absoluteEffareaCorr(EffectiveAreaCurve eain) const;

  CcdQuantumEfficiencyCurve quantumEfficiencyCurve  (double  x , 
						     double  y ) const;

  FilterTransmissionCurve filterTransmissionCurve (double theta, 
					      double phi ) const;
    
  ContaminantTransmissionCurve * contaminantTransmissionCurve () const;

  const PileupCorrector * pileupCorrector() const;

protected:

  /* Initialise patterns from DSS */
  void initPatterns( DataSubspace * dss );

  /* Check taht patterns are compatible with observing mode */
  void checkModePattern(Xmm::Mode windowMode);

  /* Set the PSF model depending on the area chosen */ 
  void setPSF( const GeoRegion & region, const RegionExtent * extent);

  /* Set factor for an incomplete extraction of a fast mode PN region */
  void setPSF_FMF(double factor);

  /* Set the SRCPOS value (for EPIC timing / burst mode) */
  void setSRCPOS(int SRCPOS);

  /* Check whether the region is an optimum ellipse or annular_ellipse */
  bool isRegionOptimumEllipse(const GeoRegion &region);

  /* Check whether a circular PSF can be used with this region */
  bool useCircularPSF(const GeoRegion &region);

  /* Find the appropriate PSF type from the input region */
  AccuracyLevel getAccuracyLevel(const GeoRegion &region);


  /* Members */
  EffectiveAreaDataServer * _effds;

  EffectiveAreaCurve _sourceposEffectiveArea; // eff. area at source pos

  InterpolatedCurve _gratingAttenuation;

  SourcePosition * _sourcePosition;
  
  CcdQuantumDataServer * _qds;
  Proxy<PatternFractionDataServer> _pfds;

  PsfModel * _psf;

  FilterDataServer * _fds;

  ContaminantDataServer * _cnds;

  PileupCorrector * _pileupCorrector;

  PositionUtils * _positionUtils;


protected:

  /* Continue initialisation of 
     derived object outside of constructor */
void  init(Xmm::Instrument instrumentID);


private:

  /* Disable default constructor */
  CalCoreData();

  /** Methods */

  void updateState();


  /** Members */

  /* DET (0.05 arcsec) to CAMCOORD2 conversion */
  double _detToCamcoord2;

  double _psfenergy;
  string _datamode;
  string _dateobs;

  const PosWCS * const _poswcs;

  vector<unsigned> _patterns;

  EventGrades _eventGrade;

  int _SRCPOS;
};
#endif
