// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by
//            pkgmaker 0.1
//            Thursday, December 11 1997
//
// Module: ConcreteArfDataServer
// Author: Hassan Siddiqui
//
// Description: Concrete implementation of ArfDataServer class
//
// Comments:
//
// Modifications: See ChangeLog
//
// $Id: ConcreteArfDataServer.cc,v 1.39 2014/04/03 13:37:51 rsaxton Exp $

// General classes
#include <cmath>

// SAS-specific classes
#include "CalServer.h"
#include "Dal.h"
#include "OalServer.h"
#include "param.h"
#include "common.h"
#include "Region.h"
#include <ASCregion.h>
#include "ExceptionErrorHandler.h"
#include "DataFilter.h"
#include "DataSubspace.h"
#include "DssBuilder.h"

// General utility functions
#include "Tracer.h"

// Local Classes
#include "ConcreteArfDataServer.h"
#include "DatafileAttributes.h"
#include "PsfModel.h"
#include "CalPsfModel.h"
#include "RatioPsfModel.h"
#include "CalMOSCoreData.h"
#include "CalPNCoreData.h"
#include "CheckExpr.h"
#include "GeneralIterator.h"
#include "TablePhotonDistribution.h"
#include "NullPhotonDistribution.h"
#include "Frame.h"
#include "ArrayPixelIterator.h"
#include "DetmapPixelIterator.h"
#include "DummyPixelIterator.h"
#include "NullPixelIterator.h"
#include "PsfmodelPixelIterator.h"
#include "NullEeCorrector.h"
#include "PointEeCorrector.h"
#include "RatioEeCorrector.h"
#include "RmfEbins.h"
#include "RegionExtent.h"
#include "BadPixelCorrector.h"
#include "NullBadPixelCorrector.h"
#include "OffCCDPixelIterator.h"
#include "RegionFunctions.h"
#include "CrossRegionArf.h"
#include "DetmapPsfModel.h"

/* Constructor */


ConcreteArfDataServer::ConcreteArfDataServer():
  _specAttributes(0),
  _rawDetmap(0),
  _coreData(0),
  _photonDistribution(0),
  _extendedsource(booleanParameter("extendedsource")),
  _modelEffectiveArea(booleanParameter("modeleffarea")),
  _modelQuantumEfficiency(booleanParameter("modelquantumeff")),
  _modelFilterTransmission(booleanParameter("modelfiltertrans")),
  _modelContamination(booleanParameter("modelcontamination")),
  _modelootcorr(booleanParameter("modelootcorr")),
  _applyxcaladjustment(booleanParameter("applyxcaladjustment")),
  _applyabsfluxcorr(booleanParameter("applyabsfluxcorr")),
  _useodf(booleanParameter("useodfatt")),
  _crossRegion(booleanParameter("crossregionarf")),
  _eeCorrector(0),
  _rfunction(0),
  _badPixCorrector(0),
  _filtBSPixelIt(0),
  _extent(0),
  _fastmode_Yfraction(1.0),
  _bstempfile("XXX") // Set to something
{
  PROC;

  /* Filtered table settings */
  _filterdss = booleanParameter("filterdss");
  _keepfilteredtable=booleanParameter("withfilteredset");
  _filteredtablename=stringParameter("filteredset");


  /* Open spectrum dataset - access=read (modify if BACKSCAL to be written) */
  string spectrumsetname = stringParameter("spectrumset");
  errstr
    << "Opening spectrumset " << spectrumsetname << "..."
    << Sas::message(Msg::App, Msg::Sparse);

  DataSet * spectrumset;

  if (booleanParameter("setbackscale")) {
     spectrumset = dataSetServer->open( spectrumsetname,
			 DataSet::Modify );
  } else {
     spectrumset = dataSetServer->open( spectrumsetname,
                         DataSet::Read );
  }

  /* Access spectrum  */
  Table * spectrumtable = spectrumset->table("SPECTRUM");

  /* Read the spectrum attributes */
  _specAttributes = new DatafileAttributes(spectrumsetname, _useodf);

  /* Get WCS information for X/Y columns */
  _poswcs = new PosWCS(spectrumset, _specAttributes->rapnt(),
              _specAttributes->decpnt(), _specAttributes->papnt(), _useodf);
  _refcoords = new CoordPair(_poswcs->wcs()->physicalRefCoords());

  /* Get observingmode */
  _windowMode = _specAttributes->getWindowMode();

  /* If the filter is bad the code exits later with a message to
     this effect. If this is a backscale run (i.e. the ARF wont be kept
     - set the filter to medium just in case */
  if (! booleanParameter("keeparfset")) {
      _specAttributes->setFilter("Medium");
  }

  /* Get filter into local variable and exit if UNKNOWN */
  _filter = _specAttributes->filter();
  if (_filter == "UNKNOWN")
  {
     errstr<<"Filter not set in file header"
            << Sas::error("BadFilterName");
  }

  /* Get Pipeline version */
  _ppsvers = _specAttributes->ppsvers();

  /* Get also DSS info */
  _dss = new DataSubspace(spectrumtable);

  /* Read useful quantities from the DSS if present */
  if (_dss->components() == 0) {
      errstr
        << "No datasubspace info found in spectrum"
        << Sas::error("DSSNotFound");
  } else {
      readDSS();
  }

  /* Get the instrument name and code */
  _instrument = _specAttributes->instrument();
  _instrumentID = _specAttributes->instrumentID();

  /* Spectrum table no longer needed. Close it */
  errstr
    << "Closing spectrumset " << spectrumsetname << "."
    << Sas::message(Msg::App, Msg::Sparse);

  dataSetServer->close(spectrumset);

  double psfenergy = realParameter("psfenergy") * 1000; // in eV.

  /* Set some CAL quantities which apply to both instruments */
  calServer.setInstrument(_instrumentID);
  calServer.state()->filter()->set(Xmm::toFilter(_filter));
  // calServer.state()->mode()->set(_windowMode);
  calServer.state()->randomize()->set(false); // Switch off randomization

  /* Now set the instrument specific CAL quantities */
  switch (_instrumentID)
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
	Sas::message(Msg::App,
		     Msg::Verbose,
		     "Creating a MOSCoreData object");
	_coreData =
	  new CalMOSCoreData( _instrumentID,
			      _specAttributes->datamode(),
			      _specAttributes->observationDate(),
			      _poswcs,
			      _filter,
			      psfenergy);
	break;
      }
    case Xmm::EPN :
      {
	Sas::message(Msg::App,
		     Msg::Verbose,
		     "Creating a PNCoreData object");
	_coreData =
	  new CalPNCoreData( _instrumentID,
			     _specAttributes->datamode(),
			     _specAttributes->observationDate(),
			     _poswcs,
			     _filter,
			     psfenergy);
	break;
      }
    default:
      errstr << "can't generate arf data for instrument of type "
	     << _instrument
	     << Sas::error("invalidInstrument");
    }

  /* Set the CCD number in the CAL - used for converting RAW to DET coords */
  calServer.state()->ccd()->set(_specAttributes->ccdNr());

  /* Check whether both DET and POS regions are defined */
  DSS::DSaxes axesdet,axespos;
  axesdet.push_back("DET");
  axespos.push_back("POS");
  if ( ! CheckExpr::check(_dss->selectExpression(axesdet)) &&
       ! CheckExpr::check(_dss->selectExpression(axespos)) ) {
     errstr<<"Spatial region defined in DETector and POS (sky) coordinates. "
             <<"arfgen cannot currently cope with this and will produce "
             <<"the wrong ARF and the wrong BACKSCAL value"
	     << Sas::warning("MixedCoords");
  }

  /* Masks are not handled very well by the DSSlib routines.
     If the spatial region consists of a single mask add a large box to the
     DSS in either sky or detector coordinates to get round this
     problem. The dummy area will be removed when the detector map is
     filtered */
  _maskFound=false;

  if ( CheckExpr::hasMask(_dss->selectExpression(axespos)) ||
       CheckExpr::hasMask(_dss->selectExpression(axesdet)) ) {

    _maskFound=true;
    std::string expression=" ";
    if ( CheckExpr::hasMask(_dss->selectExpression(axespos))) {
      expression = "(X,Y) in BOX(25921,25921,25920,25920,0)";
    } else {
      expression = "(DETX,DETY) in BOX(0,0,25920,25920,0)";
    }

    DataSubspace _dssextra = dssBuilder.getDss(expression);
    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){
      _dss->addComponent(*_dssextra.component(icomp));
    }

    errstr << "New total Expression "<<_dss->selectExpression(axespos)
         << Sas::message(Msg::App, Msg::Noisy);
  }

  /* Get the Extent of the source area */

  _TimOrBurstMode=false;
  if (_windowMode == Xmm::FAST_TIMING || _windowMode == Xmm::FAST_BURST
                                     || _windowMode == Xmm::FAST_UNCOMPRESSED
                                     || _windowMode == Xmm::FAST_COMPRESSED) {

   /* For timing and burst modes the spatial selection can be incomplete.
      In this case a pseudo-region is generated
      to handle vignetting, PSF and area issues */

     _TimOrBurstMode=true;

     _extent = addDetRegion(_coreData->varpositionUtils());
  }
  else {  // Imaging modes

     _extent = new RegionExtent(_dss, _coreData->varpositionUtils(),
               _specAttributes->datamode(),
               _specAttributes->observationDate(), _poswcs);
  }

 /* Set up a georegion with the DSS in DET or POS coords */
  _rfunction = new RegionFunctions(_coreData);
  GeoRegion dssRegion = _rfunction->region(_dss, false);

  /* Create source position object */
  bool withsourcepos = booleanParameter("withsourcepos");
  string sourceCoordType=" ";
  double sourcex=0.0;
  double sourcey=0.0;

  /* Define the position of the source from input parameters or
     DSS spatial area. */
  if (withsourcepos)
    {
      sourceCoordType = uppercase( stringParameter("sourcecoords") );
      sourcex = realParameter("sourcex");
      sourcey = realParameter("sourcey");
    }
  else
    {
      /* If the source position hasn't been specified,
         set it in detector
         coords as the centre of the source region */

      errstr
        << "Computing source position from region extent.."
        << Sas::message(Msg::App, Msg::Sparse);

      sourceCoordType = "DET";

      if (_extent->isValid())
        {
          sourcex = (_extent->xmin() + _extent->xmax()) / 2.0;
          sourcey = (_extent->ymin() + _extent->ymax()) / 2.0;
        }
      else
        {
          errstr
             << "Cannot determine source position from DSS entry"
             << Sas::error("RegionInvalid");
        }

     /* If this is a PN fast mode then need to reset the source Y position 
        to the real position of the source on the CCD */
      if (_windowMode == Xmm::FAST_TIMING || _windowMode == Xmm::FAST_BURST)
       {
        // Set the Y position of the source to be that given by SRCPOS
         sourcey = _coreData->varpositionUtils()->RawTodetxy005(
                double(32.0),double(_specAttributes->spec_SRCPOS()) ).y();
       }

    } // End of without_sourcepos block

  /* Set source position */
  _coreData->setSourcePosition(sourceCoordType, sourcex, sourcey);

  errstr
    << "Using a source position in " << sourceCoordType
    << " coordinates of x=" << sourcex
    << " y=" << sourcey
    << Sas::message(Msg::App, Msg::Sparse);

  if (sourceCoordType != "DET")
    {
      errstr
        << "This corresponds to a position in DET coordinates of "
        << "DETX=" << _coreData->sourcePosition()->xdet()
        << " DETY=" << _coreData->sourcePosition()->ydet()
        << Sas::message(Msg::App, Msg::Sparse);
    }

  errstr
    << " The source position in focal plane coords is "
    << " theta=" << _coreData->sourcePosition()->theta()
    << " phi=" << _coreData->sourcePosition()->phi()
    << Sas::message(Msg::App, Msg::Sparse);

  /* [HS] END withsourcepos logic */

  /* Initialise pattern list using selection in DSS.
     This could have been done via
     CoreData object constructor, but this approach minimises
     messy parameter propagation. */

  _coreData->initPatterns(_dss);

  /* Check whether this pattern/observing mode combination is ok */
  _coreData->checkModePattern(_windowMode);

  /* Set-up the point spread function */
  _coreData->setPSF(dssRegion, _extent);

  /* Create the detector map */
  createDetmap();

  /* Create DetmapFilter object.
     This will be used once the ArfDataServer is fully constructed. */
  _detmapfilter = new DetmapFilter;
  _filteredDetmapExists = false;

  /* Create PhotonDistribution object */
//    if (booleanParameter("modelpileup"))
//      { }
//    else
//      {
      _photonDistribution =
	new NullPhotonDistribution();
//      }

  /* Create frame object */
  _frame = new Frame( _photonDistribution );

  // Find area of source box

  errstr<<"Ignore: "<<ignoreEdgePixels()<<" "<<ignoreOutofFOV()
  << Sas::message(Msg::App, Msg::Verbose);

  /* Create Ebins object, depending on requested source */
  if (booleanParameter("withrmfset"))
    {
      _ebins = RmfEbins::ebins(stringParameter("rmfset"));
    }
  else
    {
      _ebins = _coreData->ebins();
    }

  /* If the ARF, due to flux from another region of the image is wanted
     set up an object to handle this correction. In this case all ARF
     factors are calculated at the input region except for vignetting
     and the PSF correction which are calculated from the second region */
  if (_crossRegion) {
     string cr_spectrumsetname = stringParameter("crossreg_spectrumset");
     _crossRegionArf = new CrossRegionArf(cr_spectrumsetname,
                                         _rawDetmap, positionUtils());
  }

  /* Create EeCorrector object, if source isn't extended
     and we dont want the cross-region ARF */
  if (booleanParameter("modelee")
      && (!_extendedsource)
      && (!_crossRegion)
      // && (!_TimOrBurstMode) )
       )
    {
      _eeCorrector = new PointEeCorrector( (unsigned) intParameter("eegridfactor") );

     /* Tell user which PSF is being used */
      errstr << "PSF type: "<<calServer.state()->accuracyLevel()->get()<<" will be used"
             << Sas::message(Msg::App, Msg::Sparse);
    }
//  else if (_windowMode == Xmm::FAST_BURST)
  //  {
    //  _eeCorrector = new RatioEeCorrector(_windowMode, dssRegion);
   // }
  else
    {
      _eeCorrector = new NullEeCorrector();
    }

  /* Need to set a ratio factor in the PSF for burst mode in case the
     user extracted only a part of the RAW-Y box. Use the code in
     the ratio corrector to calculate the factor */
  if (_windowMode == Xmm::FAST_BURST || _windowMode == Xmm::FAST_TIMING) {
     // const EeCorrector *RatCorrector =
       //             new RatioEeCorrector(_windowMode, dssRegion);
     // vector<double> eenergy, burst_ratio;
      //eenergy.push_back(2.0);
     // burst_ratio = RatCorrector->Fraction(eenergy);
      // Sort this
      //_burst_ratty = burst_ratio[0];
     // Sort this
      //_coreData->setPSF_FMF(burst_ratio[0]);
      _coreData->setPSF_FMF(_fastmode_Yfraction/_psfscale);
     // delete RatCorrector;
  }

  /* Set the SRCPOS vlaue for the fast mode PSF */
  _coreData->setSRCPOS(_specAttributes->spec_SRCPOS());
  
  /* Set up a constant curve objects
     in case relevant model is switched off */

  vector<double> egrid;
  egrid.push_back(0.0);
  egrid.push_back(40000.0);

  vector<double> eeff;
  eeff.push_back(1.0);
  eeff.push_back(1.0);

  _nulleacurve = new EffectiveAreaCurve(egrid, eeff);
  _nullqecurve = new CcdQuantumEfficiencyCurve(egrid, eeff);
  _nullftcurve = new FilterTransmissionCurve(egrid, eeff);
  _nullcncurve = new ContaminantTransmissionCurve(egrid, eeff);

  ENDPROC;
};

/* Destructor */
ConcreteArfDataServer:: ~ConcreteArfDataServer()
{
  PROC;

      errstr<<"Calling ConcreteArfDataServer destructor"
        << Sas::message(Msg::App, Msg::Verbose);

  delete _nullftcurve;          _nullftcurve          = 0;
  delete _nullqecurve;          _nullqecurve          = 0;
  delete _eeCorrector;          _eeCorrector          = 0;
  delete _ebins;                _ebins                = 0;
  delete _frame;                _frame                = 0;
  delete _photonDistribution;   _photonDistribution   = 0;
  delete _detmapfilter;         _detmapfilter         = 0;
  // This causes segfault when detector map is used - remove for now.
  //delete _rawDetmap;            _rawDetmap            = 0;
  delete _coreData;             _coreData             = 0;
  delete _dss;                  _dss                  = 0;
  delete _poswcs;               _poswcs               = 0;
  delete _extent;               _extent               = 0;
  delete _rfunction;            _rfunction            = 0;
  delete _badPixCorrector;      _badPixCorrector      = 0;
  delete _filtBSPixelIt; _filtBSPixelIt=0;
  // Remove the bad pixel map
  if ( dataSetServer->exists(_bstempfile))
         dataSetServer->remove(_bstempfile);  

  ENDPROC;
};

  /********************************************
   *
   * General observation/mission/instrument info
   *
   ********************************************/

Xmm::Instrument ConcreteArfDataServer::instrumentID()  const
{
  PROC;
  RETURN _instrumentID;
}

string ConcreteArfDataServer::filterName()  const
{
  PROC;
  RETURN _filter;
}
 /********************************************
   *
   * Attributes of input spectrum
   *
   ********************************************/

DatafileAttributes * ConcreteArfDataServer::specAttributes()  const
{
  PROC;
  RETURN _specAttributes;
}

/**************************
 *
 * Pointing info
 *
 **************************/

/* Source position */
double  ConcreteArfDataServer::raref()  const
{
  PROC;
  RETURN _refcoords->x();
}

double  ConcreteArfDataServer::decref() const
{
  PROC;
  RETURN _refcoords->y();
}

bool ConcreteArfDataServer::useODF() const
{
  PROC;
  RETURN _useodf;
}

bool ConcreteArfDataServer::crossRegion() const
{
  PROC;
  RETURN _crossRegion;
}

/* Method to say if pixels on the edges of CCDs are excluded in the spectrum.
This is now obsolete as flags are treated individually */
bool ConcreteArfDataServer::ignoreEdgePixels() const
{
  PROC;

      bool ignore=false;

  /* Check flag value - if zero then edge pixels are ignored. We should
     really check for the actual bit value but I can't find that info. */
      if (_flag == 0) {
         ignore=true;
      }
      errstr<<"Ignore flag set to: "<<ignore
            <<" Flag= "<<_flag
        << Sas::message(Msg::App, Msg::Verbose);

  RETURN ignore;
}

/* Method to return the event selection flag mask */
unsigned long ConcreteArfDataServer::eventFlags() const
{
  PROC;

  RETURN _flag;
}

/* Method to say if pixels outside FOV are excluded from calculations */
bool ConcreteArfDataServer::ignoreOutofFOV() const
{
  PROC;

  bool ignoreoutoffov = false;

  // All rows are included in the PN fast modes. For all other modes,
  // set flag from the input parameter
  switch (_windowMode) 
   {
     case Xmm::FAST_TIMING :
     case Xmm::FAST_BURST :
          {
            ignoreoutoffov = false;
            break;
          }
     default:
          {
            ignoreoutoffov=booleanParameter("ignoreoutoffov");
          }
   }

  RETURN ignoreoutoffov;
}

/* Method: ccdMask() - returns a mask giving the CCDs included in selection */
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => leading const removed
 */
unsigned long ConcreteArfDataServer::ccdMask() const
{
  PROC;

  RETURN _ccdMask;
}

/********************************************
 *
 * Selection info
 *
 ********************************************/
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
const DataSubspace *   ConcreteArfDataServer::dataSubspace()  const
{
  PROC;
  RETURN _dss;
}

const RegionExtent * ConcreteArfDataServer::extent()  const
{
  PROC;
  RETURN _extent;
}

/* Method to convert XY position to DETX,DETY */
SourcePosition ConcreteArfDataServer::xyToDetxy (double sourcex,
                                                double sourcey) const
{
  PROC;

     errstr<<"Entered xyToDetxy: "<<sourcex<<" "<<sourcey
    << Sas::message(Msg::App, Msg::Verbose);

 /* This uses a temporary positionUtils object. This must change and
    is pending a restructuring */

  PositionUtils * testpos = new PositionUtils(_instrumentID);

  SourcePosition s = testpos->sourcePosition( "POS",
                                                     _specAttributes->datamode(),
                                                     _specAttributes->observationDate(),
                                                     sourcex,
                                                     sourcey,
                                                     _poswcs);
  RETURN s;
}


/********************************************
 *
 * Ebins accessor
 *
 ********************************************/
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
const BinnedAxis *     ConcreteArfDataServer::ebins()   const
{
  PROC;
  RETURN _ebins;
}

/********************************************
 *
 * Source Position
 *
 ********************************************/
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
const SourcePosition *   ConcreteArfDataServer::sourcePosition() const
{
  PROC;
  RETURN _coreData->sourcePosition();
}


/********************************************
 *
 * Misc
 *
 ********************************************/

bool ConcreteArfDataServer::filterDSS ()    const
{
  PROC;
  RETURN _filterdss;
}


const string ConcreteArfDataServer::filteredTableName ()    const
{
  PROC;
  RETURN _filteredtablename;
}

bool ConcreteArfDataServer::keepFilteredTable ()    const
{
  PROC;
  RETURN _keepfilteredtable;
}

bool ConcreteArfDataServer::isExtendedSource() const
{ PROC; RETURN _extendedsource;}

/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
const PhotonDistribution *
ConcreteArfDataServer::photonDistibution() const
{
  PROC;
  RETURN _photonDistribution;
}
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
const Frame *   ConcreteArfDataServer::frame() const
{
  PROC;
  RETURN _frame;
}
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
PsfModel *   ConcreteArfDataServer::psf() const
{
  PROC;
  RETURN _coreData->psf();
}
/* detmapPixelIterator:  Filter the detector map */
 PixelIterator *   ConcreteArfDataServer::detmapPixelIterator()
{

  PROC;

  /* Ideally, the filtered detmap would be created
     during construction of the arfDataServer, but the process unfortunately needs
     other quantities from the server. */

    if (! _filteredDetmapExists) {

  /* If we have added a pseudo-region to handle X/Y image masks then
     remove it during the filtering stage */
       if (_maskFound) {
          _dss->deleteComponent(_dss->components()-1);
       }

       _filteredDetmap = _detmapfilter->filteredMap(_rawDetmap, 
                          filteredTableName()+":EVENTS", _specAttributes, _dss, false);
       _filteredDetmapExists = true;

     // Add the pseudo-component back in
 //      if (_maskFound) {
 //      DScomp * dscompsave = _dss->component(_dss->components()-1);
  //         _dss->addComponent(*dscompsave);
   //    }
    }

  RETURN _filteredDetmap;
}

/* Set up the bad pixel correctors */
void ConcreteArfDataServer::set_bpcorrectors()
{

  PROC;

   /* If a pixel mask is being used then recalculate the sourceboxarea
      using the detector map */
   /*
    if (_maskFound)
    {
     // Count the number of non-zero pixels in the detector map
       PixelElement pixel = _filteredDetmap->normalisedItem();
       _sourceBoxArea=INT(_filteredDetmap->size()
                             *pixel.xWidth()*pixel.yWidth());

       errstr<<"Recalculated source box size: "<<_sourceBoxArea
       << Sas::message(Msg::App, Msg::Verbose);
    }
    else {
 */
 /* Make a map of all the possible pixels with a resolution of 1 arcsecond */
  //PixelIterator * BSPixelIt = new DummyPixelIterator(-21000.0,21000.0,2100,
   //                                  -21000.0,21000.0,2100);

   /* Get the bad pixel map definition parameter (arcsecs) */
    float pixres = 0.0;
    if (booleanParameter("withbadpixres"))
       pixres = realParameter("badpixelresolution");
    else {
      // If we are performing the PSF correction set to 0.5"
      // or 0.25" for Timing mode
       if (!_extendedsource) {
          if (_windowMode==Xmm::FAST_TIMING ||
              _windowMode==Xmm::FAST_BURST) {              
              pixres = 0.25;
          } else {
              pixres = 0.5;
          }
       } else { // otherwise
          pixres = 2.0;
       }
    }

   /* Set up a map for bad pixel checking. Initially with a size which
      encompasses the whole region and then filtered to the region itself. 
      Limit to a 42000 width to avoid poor coord conversion and long
      execution times. */
    double xstart = max((_extent->xmin() - 50.0), -21000.0);
    double ystart = max((_extent->ymin() - 50.0), -21000.0);
    double xend = min((_extent->xmax() + 50.0), 21000.0);
    double yend = min((_extent->ymax() + 50.0), 21000.0);
    int npixx = INT( (xend - xstart)  /
                     (20.0 * pixres) );
    int npixy = INT( (yend - ystart) /
                     (20.0 * pixres) );
    
   // int npix = INT(42000.0 / (20.0 * pixres));
   // PixelIterator * BSPixelIt = new DummyPixelIterator(-21000.0,20999.0,npix,
     //                                  -21000.0,20999.0,npix);
   // !!!!!! RDS CHANGE - SPR 7186 !!!!!!! //
    //PixelIterator * BSPixelIt = new DummyPixelIterator(xstart,
      //              (xstart + pixres*20.0*npixx-1), npixx,
        //            ystart, (ystart + pixres*20.0*npixy-1), npixy);
    PixelIterator * BSPixelIt = new DummyPixelIterator(xstart,
                    (xstart + pixres*20.0*npixx), npixx,
                    ystart, (ystart + pixres*20.0*npixy), npixy);

   /* Filter this map with the selection expression - use unique tempfile */
    //_bstempfile = Filename::make_temp_filename("BSMAP"); 
    _bstempfile = "BSMAP.ds";
    string bstempfile = _bstempfile + ":EVENTS";
    DetmapFilter * dmfilter = new DetmapFilter();
    _filtBSPixelIt = dmfilter->filteredMap(BSPixelIt,
                        bstempfile, _specAttributes, _dss, false);

 /* Find the size of the filtered map */
  //_sourceBoxArea = INT(_filtBSPixelIt->size()*20*20*pixres*pixres);
  _sourceBoxArea = INT(20.0*20.0*pixres*pixres) * _filtBSPixelIt->size();

  errstr<<"Source box size: "<<_sourceBoxArea
  << Sas::message(Msg::App, Msg::Verbose);

     //delete dmfilter; dmfilter=0;
  //  }

  /* Create a bad pixel corrector if required */
    if (booleanParameter("withbadpixcorr"))
    {

      /* Create a correction vector for the bad pixels depending on
         whether the source is deemed to be point-like or extended */
      if (_extendedsource) {

        // Use the detector map if required
         if ( stringParameter("badpixmaptype") =="dataset" )
            _bp_psf = new DetmapPsfModel( _filteredDetmap );
         else
            _bp_psf = new RatioPsfModel( _sourceBoxArea );

      } else { // (point source: so use the same PSF as for modelee corrs.)

         _bp_psf = _coreData->psf();
      }

      _badPixCorrector = new BadPixelCorrector(
                            stringParameter("badpixlocation"),
                            (unsigned) intParameter("eegridfactor"),
                            _bp_psf, _flag, _ccdMask );

    }
  else {
      _badPixCorrector = new NullBadPixelCorrector();
  }
}

/* Provide a pixel iterator to evaluate Off-axis pixels etc. */
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
PixelIterator *   ConcreteArfDataServer::filteredBSPixelIt()
{
   RETURN _filtBSPixelIt;
}

/* Provide a pixel iterator to evaluate the PSF function*/
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
PixelIterator * ConcreteArfDataServer::PsfPixelIterator()
{

  PROC;

  /* If this is a circular PSF function then we just need a dummy one
     pixel iterator. Otherwise integrate over the detector map */

     errstr<<"Acclevel is: "<<calServer.state()->accuracyLevel()->get()
    << Sas::message(Msg::App, Msg::Verbose);

    if ( calServer.state()->accuracyLevel()->get() == EXTENDED ) 
    {
     errstr<<"Acclevel dum"
    << Sas::message(Msg::App, Msg::Verbose);
       //RETURN (new DummyPixelIterator( 1.0, 2.0, 1, 1.0, 2.0, 1 ));
       RETURN filteredBSPixelIt();  
    }
    else if (calServer.state()->accuracyLevel()->get() == ELLBETA) {
      // Return iterator for the source area at a resolution of
      // {badpixelresolution} arcsecs per bin
       RETURN filteredBSPixelIt();  
    }
    else {
     errstr<<"Acclevel detmap"
    << Sas::message(Msg::App, Msg::Verbose);
       RETURN detmapPixelIterator();
    }
}

double ConcreteArfDataServer::framePeriod()    const
{
  PROC;
  RETURN _coreData->framePeriod();
}

bool ConcreteArfDataServer::setCcd( const CamCoords & camcoord)
{
  PROC;
  RETURN _coreData->setCcd(camcoord);
}


/* Convert from CAMCOORD2 to PIXCOORD */
const PixCoords ConcreteArfDataServer::camCoord2ToPixCoord( const CamCoords & camcoords ) const
{
      PROC;
      RETURN _coreData->camCoord2ToPixCoord(camcoords);
}

/* Return the positionUtils object */
const PositionUtils * ConcreteArfDataServer::positionUtils() const
{ PROC;
   RETURN _coreData->positionUtils();
}

/* Convert from DETX/DETY (0.05 arcsec units) to camcoord2 */
const CamCoords ConcreteArfDataServer::detxy005ToCamCoord2(int detx_005,
							   int dety_005) const
{
  PROC;
  RETURN _coreData->detxy005ToCamCoord2( detx_005,
					 dety_005 );
}

/* Convert from CAMCOORD2 to TELCOORD */
const TelCoords ConcreteArfDataServer::camCoord2ToTelCoord( const CamCoords & camcoords ) const
{
  PROC;
  RETURN _coreData->camCoord2ToTelCoord(camcoords);
}



/***********************
 *  Arf quantity routines
 ***********************/

/* These methods can be turned on/off by user */
EffectiveAreaCurve ConcreteArfDataServer::effectiveAreaCurve ( double theta,
                double phi ) const
{
  PROC;
  if (!_modelEffectiveArea) ALTRETURNV( *_nulleacurve );

  //double area;
  EffectiveAreaCurve eacurve;

  /* If cross region ARF wanted find the product of the
  effective area (including vignetting) at the flux region and the
  PSF contribution in the source region */
  if (_crossRegion)
     RETURN _crossRegionArf->psfVignet(theta, phi);

/* For point sources return the effective area at the source position */
  if (! _extendedsource) {
     eacurve=_coreData->effectiveArea();
     errstr<<"Calculating eff. area at source pos"
           << Sas::message(Msg::App, Msg::Verbose);
  }
/* For extended sources return effective area at the input position */
  else {
     eacurve=_coreData->effectiveArea(theta, phi);
  }

/* Apply the mirror effective area correction array if requested */
  if (_applyxcaladjustment) {
     eacurve = _coreData->mirrorAdjustment(eacurve);
  }

/* Apply the effective area correction designed to make EPIC fluxes agree
   with those of other missions */
  if (_applyabsfluxcorr) {
     eacurve = _coreData->absoluteEffareaCorr(eacurve);
  }

  RETURN eacurve;
}

CcdQuantumEfficiencyCurve
ConcreteArfDataServer::quantumEfficiencyCurve   ( double  x ,
						  double  y  ) const
{
  PROC;
  if (!_modelQuantumEfficiency)
    {
      ALTRETURNV( *_nullqecurve );
    }

  errstr
    << "Getting QE curve for position "
    << x << "," << y << "..."
    << Sas::message(Msg::App, Msg::Noisy);

  CcdQuantumEfficiencyCurve qecurve =       
    _coreData->quantumEfficiencyCurve(x,y);

  RETURN qecurve;
}

FilterTransmissionCurve
ConcreteArfDataServer::filterTransmissionCurve   ( double  theta ,
						   double  phi  ) const
{
  PROC;
  if (!_modelFilterTransmission)
    {
      ALTRETURNV( *_nullftcurve );
    }

  FilterTransmissionCurve ftcurve =
    _coreData->filterTransmissionCurve(theta, phi);

  RETURN ftcurve;
}

ContaminantTransmissionCurve *
ConcreteArfDataServer::contaminantTransmissionCurve() const
{
  PROC;
  if (!_modelContamination)
    {
      ALTRETURNV( _nullcncurve );
    }

  RETURN _coreData->contaminantTransmissionCurve();
}


double ConcreteArfDataServer::pileupCorrection   (double  energy,
						  double  pixelFlux) const
{
  PROC;

  double corrn =  _coreData->
    pileupCorrector()->
    correction(energy, pixelFlux);

  RETURN corrn;
}


/* Method to return the out-of-time correction value */
double ConcreteArfDataServer::ootCorrection ( ) const
{
  PROC;

  if (!_modelootcorr)
    {
      ALTRETURNV( 1.0 );
    }

  double ootfact=1.0;

/* Get the pipeline date from the ppsvers string */
  string ppsdate;

  if (_ppsvers != "NotSet" && _ppsvers.find("/") != string::npos
     && _ppsvers.find(".") != string::npos)
  {
     ppsdate = _ppsvers.substr(_ppsvers.find("/"),_ppsvers.find("."));
  }
  else {
     ppsdate = "99999999";

     /* Remove this warning as it is no longer relevant */
     //errstr<<"No PPS version present in header "
      //       <<"not applying OOT correction"
	   //    << Sas::warning("NoPPSVersion");
  }

/*  The algorithm is that the out-of-time effect is zero unless this
    is a PN observation in either FF or EFF mode taken from a spectrum
    produced with an earlier pipeline version than 18-June-2001.
    MOS oot factors are included in the exposure time always. */

  if (_instrumentID == Xmm::EPN && ppsdate < "20010618" )
    {
      switch (_windowMode) {

        case Xmm::PRIME_FULL_WINDOW:
          {
            ootfact = 0.938;
            break;
          }
        case Xmm::PRIME_FULL_WINDOW_EXTENDED:
          {
            ootfact = 0.977;
            break;
          }
        case Xmm::PRIME_SMALL_WINDOW:
          {
            ootfact = 0.989;
            break;
          }
        default:
            ootfact = 1.0;
      }
    } else {
       ootfact = 1.0;
    }

   RETURN ootfact;
}

/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => qualifiers removed
 * const XXXXX * const yyy()   const;
 * const XXXXX *       yyy()   const;
 */
const EeCorrector *  ConcreteArfDataServer::eeCorrector () const
{
  PROC;
  RETURN _eeCorrector;
}
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => qualifiers removed
 * const XXXXX * const yyy()   const;
 * const XXXXX *       yyy()   const;
 */
const BpCorrector *  ConcreteArfDataServer::badPixCorrector () const
{
  PROC;
  RETURN _badPixCorrector;
}

/* Method to return the backscale value */
int ConcreteArfDataServer::backScale() const
{
  PROC;

   /* return the number of detector pixels in the source area minus
      the bad pixels and the chip gaps */
  int totarea = _sourceBoxArea * _fastmode_Yfraction - 
                                      _badPixCorrector->totalPixels();
    errstr<<"Source box area "<<_sourceBoxArea*_fastmode_Yfraction<<" "
          <<_fastmode_Yfraction<<" "
          <<"totpix "<< _badPixCorrector->totalPixels()
          <<"totarea "<<totarea
          << Sas::message(Msg::App, Msg::Verbose);

  RETURN totarea;
}

int ConcreteArfDataServer::sourceBoxArea() const
{
  PROC;

  RETURN _sourceBoxArea;
}

// Checks if detector map is from same observation and instrument as spectrum
bool ConcreteArfDataServer::standardDetMap(const string & arrayname)
{
  if (! _specAttributes)
     errstr<<"Spectral file attributes not yet defined"
           <<Sas::error("SpecAttErr");

  bool standard=true;

  DataSet* set = dataSetServer->open( arrayname, DataSet::Read );

  // Check if map is in detector coords and comes from same observation and
  // instrument as the spectrum
  if (set->hasAttribute("DATE-OBS") && set->hasAttribute("INSTRUME")) {
   if (set->attribute("DATE-OBS")->asString() != _specAttributes->observationDate())
      standard=false;
   if (set->attribute("INSTRUME")->asString() != _specAttributes->instrument())
      standard=false;
  } else {
      standard=false;
  }

  string coordtype = "NotSet";
  if (set->hasAttribute("CTYPE1"))
  {
     coordtype = set->attribute("CTYPE1")->asString();
  } else { // Try opening as an array and looking in this header
      if (arrayname.find(":") != string::npos)
        {
           Array *delArr = set->array(arrayname);

           if (delArr->hasAttribute("CTYPE1")) {
             coordtype = delArr->attribute("CTYPE1")->asString();
           }
           //delete delArr; delArr=0;
        }
  }
  if (coordtype.find("DET") == string::npos)
     standard=false;

  dataSetServer->close(set); 

  return standard;
}


void ConcreteArfDataServer::createDetmap()
{
  /* Define variables to hold detmap bounds.
     The source position will be checked
     later to see if it is within these bounds
  */
  double srcxpos = _coreData->sourcePosition()->xdet();
  double srcypos = _coreData->sourcePosition()->ydet();

  /* Construct PixelIterator object */
  string detmaptype = stringParameter("detmaptype");

  if ( detmaptype =="dataset" )
    {
      errstr << "Constructing a DATASET-based detector map."
	<< Sas::message(Msg::App, Msg::Sparse);

     /* Add colon to arrayname if necessary*/
      string arrayname = stringParameter("detmaparray");
      if (arrayname.find(":") == string::npos) arrayname += ":";

     /* If the array is from the same observation and instrument as the spectrum
      * and is already in detector coords then create a simple array iterator */
      if (standardDetMap(arrayname)) {
         _rawDetmap = new ArrayPixelIterator(arrayname);

      } else {

       // Create an iterator which converts coordinates from the array 
       // into detector coords in the spectrum coord frame */
       // HORRIBLE USE OF Param 
         _rawDetmap = new DetmapPixelIterator(arrayname, 
                                  stringParameter("spectrumset"));
      }
      
      /* Now need to access size */
      unsigned xsize = _rawDetmap->xbins();
      unsigned ysize = _rawDetmap->size()/xsize;

      CoordPair rawboundsmin = _rawDetmap->physXYmin();
      CoordPair rawboundsmax = _rawDetmap->physXYmax();

      errstr
        << "The dataset-based detmap bounds corresponds to "
        << " xmin="  << rawboundsmin.x()
        << " xmax="  << rawboundsmax.x()
        << " xbins=" << xsize
        << " ymin="  << rawboundsmin.y()
        << " ymax="  << rawboundsmax.y()
        << " ybins=" << ysize
        << Sas::message(Msg::App, Msg::Sparse);

      /* Check whether the extent of the regions are larger
	 than that of the detector map */
      if(_extent->isValid())
	{
          _extent->checkBounds( rawboundsmin.x(), rawboundsmax.x(),
                                rawboundsmin.y(), rawboundsmax.y());
	}
    }
  else
    {
      /* PSF or FLAT detector map */

      // Find the bounds of the detector map
      double xmin = 0, xmax = 0, ymin = 0, ymax = 0;
      if (booleanParameter("withdetbounds"))
	  {
	    errstr
	      << "Taking detector map bounds from the parameters "
              << "detxoffset and detyoffset ..."
	      << Sas::message(Msg::App, Msg::Sparse);

	    double xoffset = realParameter("detxoffset");
	    double yoffset = realParameter("detyoffset");

	    xmin = srcxpos-xoffset;
	    xmax = srcxpos+xoffset;
	    ymin = srcypos-yoffset;
	    ymax = srcypos+yoffset;
	  }
      else
	{
	  errstr
	    << "Computing detector map bounds from the region selection "
            << "information stored in the spectrum ..."
	    << Sas::message(Msg::App, Msg::Sparse);

	  if (_extent->isValid())
	    {
	      xmin = _extent->xmin();
	      xmax = _extent->xmax();
	      ymin = _extent->ymin();
	      ymax = _extent->ymax();
	    }
	  else
	    {
	      errstr
		<< "Cannot determine bounds for detector map from region extent."
		<< "Either check DSS information in spectrum,"
		<< "or set the bounds explicitly by setting the parameter --withdetbounds to true,"
		<< " and entering appropriate values for --detxoffset and --detyoffset."
		<< Sas::error("RegionInvalid");
	    }
	}

     /* Find number of detector map pixels from user-supplied values */
      unsigned xbins=0, ybins=0;
      if (booleanParameter("withdetbins"))
         {
           xbins = (unsigned) intParameter("detxbins");
           ybins = (unsigned) intParameter("detybins");
         }
      else // or determine the best number of bins to use from region
         {
          // Use 4 bins per arcminute with minimum of 5 and max of 80 bins
           xbins = max(5, min(80, INT((4 * abs(xmax - xmin) / 1200)))); 
           ybins = max(5, min(80, INT((4 * abs(ymax - ymin) / 1200)))); 
         }

      errstr<< "detector map binning will be: "
            << "xbins: "<<xbins <<" ybins: "<<ybins
            << Sas::message(Msg::App, Msg::Sparse);

     /* Make a PSF or FLAT detmap */
      if (detmaptype=="psf")
	{
	   _rawDetmap =
	       new PsfmodelPixelIterator( xmin, xmax, xbins,
				          ymin, ymax, ybins );
	}
      else
	{
	  _rawDetmap =
	       new DummyPixelIterator( xmin, xmax, xbins,
                                          ymin, ymax, ybins );
	}

      errstr << "Constructing a " <<detmaptype<< " detector map with "
             << " xmin="  << xmin << " xmax="  << xmax
             << " xbins=" << xbins << " ymin="  << ymin
             << " ymax="  << ymax << " ybins=" << ybins
             << Sas::message(Msg::App, Msg::Sparse);
    }

  /* Check that source position is within bounds of map */
  if (!_rawDetmap)
    {
      errstr<<"Anomaly in ConcreteArfDataServer - raise SPR"
          << Sas::error("ARFObjError");
    }

  ENDPROC;
}

/* addDetRegion - adds a pseudo spatial region in detector coords to the dss */
RegionExtent * ConcreteArfDataServer::addDetRegion(PositionUtils * putils)
{
  PROC;

  // Find the RAW-X and RAW-Y limits for this instrument
  double lowlim_rawy=0, upplim_rawy=0, lowlim_rawx=0, upplim_rawx=0;
  _psfscale=1.0; // Need an extra number to scale PSF for Burst mode later

  switch (_instrumentID)
       {
       case Xmm::EMOS1 :
       case Xmm::EMOS2 :
         {
           lowlim_rawx=250, upplim_rawx=350;
           lowlim_rawy=1, upplim_rawy=600;
           break;
         }
       case Xmm::EPN :
         {
           lowlim_rawx=1, upplim_rawx=64;
           lowlim_rawy = 1;

           if (_windowMode==Xmm::FAST_TIMING) {
              upplim_rawy = 200.0;
           } else if (_windowMode==Xmm::FAST_BURST) {    
              upplim_rawy = 200.0;
              _psfscale=180.0/200.0;
           }
           break;
         }
       default:
        errstr<<"Instrument "<<_instrumentID<<" not recognised"
              << Sas::error("invalidInstrument");
       }

  /* Get the RAW coord selection for debug */
  DSS::DSaxes axes;
  axes.push_back("RAW");
  string projregionexpr = _dss->selectExpression(axes);
  errstr << "Expression "<<projregionexpr
         << Sas::message(Msg::App, Msg::Verbose);

  // Find the extents of the DSS in RAW-X
  double lower_rawx=0, lower_rawy=0, upper_rawx=0, upper_rawy=0;
  RangeBound lowerB_rawx, lowerB_rawy, upperB_rawx, upperB_rawy;

  list<Range<double> > rangeX = _dss->ranges("RAWX");
  Range<double> extent_rawy = _dss->extent("RAWY");
  extent_rawy.values(lower_rawy, upper_rawy, lowerB_rawy, upperB_rawy);

  // Rest the Y range to defaults if unbounded
  if ( lowerB_rawy == Unbounded ) lower_rawy = lowlim_rawy;
  if ( upperB_rawy == Unbounded ) upper_rawy = upplim_rawy*_psfscale;

  // Here we set the actual Y range used to extract the data. This
  // will be used to scale the PSF and backscale values later.
  // We use the full 1-200 range for the detector region so that the
  // PSF can be correctly modelled. However, the full PSF is contained within
  // just 180 pixels for BURST mode
  _fastmode_Yfraction = (upper_rawy - lower_rawy + 1) / 
                           (upplim_rawy - lowlim_rawy +1);

  errstr << "Y fraction: "<<_fastmode_Yfraction
         << Sas::message(Msg::App, Msg::Verbose);

  /* Make proper 2-d regions from ranges */
  if (rangeX.begin() != rangeX.end())
  {

   /* Loop over each RAWX range - assume RAWY is constant */
   for ( list<Range<double> >::iterator i = rangeX.begin();
     i != rangeX.end(); ++i ) {

     (*i).values(lower_rawx, upper_rawx, lowerB_rawx, upperB_rawx);

    // Set the RAWX range if unbounded
     if ( lowerB_rawx == Unbounded ) lower_rawx = lowlim_rawx;
     if ( upperB_rawx == Unbounded ) upper_rawx = upplim_rawx;

    /* If either extent is unbounded then we need to take defaults to
       convert the RAW X/Y extent into a 2-d box */

  //   if ( lowerB_rawx == Unbounded || upperB_rawx == Unbounded ||
    //  PUT BACK  lowerB_rawy == Unbounded || upperB_rawy == Unbounded ) {

    // Now set the Y range to the maximum pixels on the Chip
     lower_rawy = lowlim_rawy, upper_rawy = upplim_rawy;

   /* Find the size of region taking into account that more than one
      RAW coord range may have been selected !!! - fix here for SPR-2595 */
    double xcen = (lower_rawx + upper_rawx) / 2.0;
    double ycen = (lower_rawy + upper_rawy) / 2.0;
    double x_hwid = abs(xcen - lower_rawx) +0.5;   // Dealing with integer pix
    double y_hwid = abs(ycen - lower_rawy) +0.5;

   // Create an expression from these coords
    std::string expression;
    expression="(RAWX,RAWY) in BOX("+toString(xcen)+","+toString(ycen)+
               ","+toString(x_hwid)+","+toString(y_hwid)+",0)";

    errstr<<"From raw X: "<<lower_rawx<<" "<<upper_rawx
              <<" raw Y: "<<lower_rawy<<" "<<upper_rawy<<" got: "
              <<expression
              << Sas::message(Msg::App, Msg::Verbose);

   /* Add a new component to the DSS for this RAWX/Y extraction */
    DataSubspace _dssextra = dssBuilder.getDss(expression);

    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){

      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

      DScomp * dscomp = _dssextra.component(icomp);

     /* Loop over each filter in the new component */
      for (unsigned f = 0; f < dscomp->filters(); ++f) {

       /* Add this filter to each component of the original DSS */
        for(unsigned idc=0; idc <_dss->components(); idc++){

           DScomp * oldcomp = _dss->component(idc);

           oldcomp->addFilter(*dscomp->filter(f));
        }
      }
    }

    errstr << "New total Expression "<<_dss->selectExpression(axes)
         << Sas::message(Msg::App, Msg::Verbose);

 //  } // End of if rawx/y unbounded block - PUT BACK
  } // End of loop over ranges
 } else 
 {  // No RAW ranges selected but full regions

   /* Add a region covering the full Y range and the selected X range.
      This is done to handle the PSF calculation in burst mode where
      a fraction of the Y range is often used to avoid leakage of the 
      point source */

    // Set a region functions object from the coredata
   RegionFunctions * rfunction2 = new RegionFunctions(_coreData);
   GeoRegion inRegion = rfunction2->region(_dss, true);

 /* Loop over each component in the input region */
   for (unsigned comp=0; comp<inRegion.size(); comp++) {

      double xmin=1000.0, xmax=0.0;

     /* Loop over each shape in this component */
      for (unsigned element=0; element<inRegion[comp].size(); element++) {

       /* Get X RAW coord values for the shape */
        std::vector<float> X = inRegion[comp][element]->pointsX();
        double rawx_width = inRegion[comp][element]->radii()[0];

        xmin = min(xmin,(X[0] - rawx_width/2.0));
        xmax = max(xmax,(X[0] + rawx_width/2.0));
      }
  //  }

   /* Find the size of region taking into account that more than one
      RAW coord range may have been selected !!! - fix here for SPR-2595 */
    double xcen = (xmin + xmax) / 2.0;
    double ycen = (lowlim_rawy + upplim_rawy) / 2.0;
    double x_hwid = abs(xcen - xmin);   // Dealing with integer pix
    double y_hwid = abs(ycen - lowlim_rawy) +0.5;

   // Create an expression from these coords
    std::string expression;
    expression="(RAWX,RAWY) in BOX("+toString(xcen)+","+toString(ycen)+
               ","+toString(x_hwid)+","+toString(y_hwid)+",0)";

    errstr<<"From raw X: "<<xmin<<" "<<xmax
              <<" raw Y: "<<lower_rawy<<" "<<upper_rawy<<" got: "
              <<expression
              << Sas::message(Msg::App, Msg::Verbose);

   /* Add a new component to the DSS for this RAWX/Y extraction */
    DataSubspace _dssextra = dssBuilder.getDss(expression);

    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){

      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

      DScomp * dscomp = _dssextra.component(icomp);

     /* Loop over each filter in the new component */
      for (unsigned f = 0; f < dscomp->filters(); ++f) {

       /* Add this filter to each component of the original DSS */
       for(unsigned idc=0; idc <_dss->components(); idc++){

         DScomp * oldcomp = _dss->component(idc);

         oldcomp->addFilter(*dscomp->filter(f));
       }
      }
    }

    errstr << "Newer total Expression "<<_dss->selectExpression(axes)
         << Sas::message(Msg::App, Msg::Verbose);
   } // REMOVE !!!
  } // End of No ranges selected block

  // Set a region functions object from the coredata
  RegionFunctions * rfunction = new RegionFunctions(_coreData);

 /* Convert all RAW X/Y filters into DET X/Y */
  rfunction->convRawToDet(_dss);

 /* Return the converted datasubspace as a region extent object */
  RegionExtent * regext = new RegionExtent(rfunction->detDSS(), putils,
               _specAttributes->datamode(), _specAttributes->observationDate(),
               _poswcs);

  RETURN regext;
}

/* Method readDSS() - reads some values from the data subspace */
void ConcreteArfDataServer::readDSS()
{
  PROC;

    DSS::DSaxes axestest;
    axestest.push_back("FLAG");

     _flag=0xffffffff;
     _ccdMask=0;

   /* Read the FLAG value if any from the DSS */
     for (unsigned i  = 0; i < _dss->components(); ++i) {

       DScomp * dscomp = _dss->component(i);

     /* Loop over each filter in this component */
       for (unsigned f = 0; f < dscomp->filters(); ++f) {

          DSfilter * dsfilt = dscomp->filter(f);

        /* Try and find a component which has a FLAG filter */
          if (dsfilt->name() == "FLAG" || dsfilt->name() == "CCDNR")
          {

            errstr<<"filter: "<<f<<" "<<i<<" "<<dsfilt<<" "
                  <<dsfilt->name()<<" "<<dsfilt->type()
                << Sas::message(Msg::App, Msg::Noisy);

           /* Have to handle masks and ranges separately */
            if (dsfilt->type() == DataSubspace::BitMaskT) {
               //vector<unsigned long> vecflag = dsfilt->offBitMasks();
               vector<uint32_t> vecflag = dsfilt->offBitMasks();

               for ( unsigned vf = 0; vf < vecflag.size(); ++vf ) {
                   _flag &= ~vecflag[vf];
               }
            }
            else if (dsfilt->type() == DataSubspace::RangeT) {

               list <Range<double> > ranges = dsfilt->ranges();

               double upper, lower;
               RangeBound upperB, lowerB;

             /* Treat this mask as a range, though it may be a single value */
               for ( list<Range<double> >::iterator it = ranges.begin();
                  it != ranges.end(); ++it )
               {
                  (*it).values(lower, upper, lowerB, upperB);

              // Process this range if bounded
                if (lowerB != Unbounded && upperB != Unbounded)
                {

                  unsigned long ulower = (unsigned long) lower,
                             uupper = (unsigned long) upper;

                  if (dsfilt->name() == "FLAG")
                    _flag &= ulower;
                  else // CCD mask set - "OR" these CCDs with current mask
                  {
                    for (unsigned long c=ulower; c<=uupper; c++)
                       { _ccdMask |= 1<<c; }
                  }
                } // end of if bounded
              } // end of ranges loop

            } // End of if range block

         }
         errstr<<"Setting FLAG BIT mask "<<_flag
	           <<" CCD mask "<<_ccdMask
               << Sas::message(Msg::App, Msg::Noisy);
       }

     }
    /* If no CCD selection has been made assume that all CCDs are
       included */
     if (! _ccdMask) {
        for (unsigned long c=1; c<=12; c++) {
                  _ccdMask += 1<<c;
        }
        errstr<<"Defaulting CCD mask "<<_ccdMask
              << Sas::message(Msg::App, Msg::Verbose);
     }

  ENDPROC;
}

/* Method PolyRegion() - returns the source region as polygons */
GeoRegion ConcreteArfDataServer::PolyRegion() const
{
  PROC;

  RETURN _rfunction->regionToPoly(_rfunction->region(_dss, false));
}

